@@ -71,13 +71,23 @@ struct raft_buffer
     size_t len; /* Length of the buffer. */
 };
 
+/**
+ * A type for storing unknown bools.
+ */
+typedef enum {
+    raft_tribool_unknown,
+    raft_tribool_true,
+    raft_tribool_false,
+} raft_tribool;
+#define TO_RAFT_TRIBOOL(b)  ((b) ? raft_tribool_true : raft_tribool_false)
+
 /**
  * Server role codes.
  */
 
 #define RAFT_STANDBY 0 /* Replicate log, does not participate in quorum. */
 #define RAFT_VOTER 1   /* Replicate log, does participate in quorum. */
-#define RAFT_SPARE 2    /* Does not replicate log, or participate in quorum. */
+#define RAFT_SPARE 2   /* Does not replicate log, or participate in quorum. */
 
 /**
  * Hold information about a single server in the cluster configuration.
@@ -252,8 +262,9 @@ struct raft_request_vote
  */
 struct raft_request_vote_result
 {
-    raft_term term;    /* Receiver's current term (candidate updates itself). */
-    bool vote_granted; /* True means candidate received vote. */
+    raft_term term;        /* Receiver's current term (candidate updates itself). */
+    bool vote_granted;     /* True means candidate received vote. */
+    raft_tribool pre_vote; /* The response to a pre-vote RequestVote or not. */
 };
 
 /**
@@ -416,6 +427,11 @@ struct raft_tracer
      */
     void *impl;
 
+    /**
+     * Whether this tracer should emit messages.
+     */
+    bool enabled;
+
     /**
      * Emit the given trace message, possibly decorating it with the provided
      * metadata.
@@ -485,6 +501,31 @@ struct raft_io
     int (*random)(struct raft_io *io, int min, int max);
 };
 
+/*
+ * version 1:
+ * struct raft_fsm
+ * {
+ *     int version;
+ *     void *data;
+ *     int (*apply)(struct raft_fsm *fsm,
+ *                  const struct raft_buffer *buf,
+ *                  void **result);
+ *     int (*snapshot)(struct raft_fsm *fsm,
+ *                     struct raft_buffer *bufs[],
+ *                     unsigned *n_bufs);
+ *     int (*restore)(struct raft_fsm *fsm, struct raft_buffer *buf);
+ * };
+ *
+ * version 2:
+ * introduces `snapshot_finalize`, when this method is not NULL, it will
+ * always run after a successful call to `snapshot`, whether the snapshot has
+ * been successfully written to disk or not. If it is set, raft will
+ * assume no ownership of any of the `raft_buffer`s and the responsibility to
+ * clean up lies with the user of raft.
+ * `snapshot_finalize` can be used to e.g. release a lock that was taken during
+ * a call to `snapshot`. Until `snapshot_finalize` is called, raft can access
+ * the data contained in the `raft_buffer`s.
+ */
 struct raft_fsm
 {
     int version;
@@ -496,6 +537,9 @@ struct raft_fsm
                     struct raft_buffer *bufs[],
                     unsigned *n_bufs);
     int (*restore)(struct raft_fsm *fsm, struct raft_buffer *buf);
+    int (*snapshot_finalize)(struct raft_fsm *fsm,
+                             struct raft_buffer *bufs[],
+                             unsigned *n_bufs);
 };
 
 /**
@@ -508,12 +552,13 @@ enum { RAFT_UNAVAILABLE, RAFT_FOLLOWER, RAFT_CANDIDATE, RAFT_LEADER };
  */
 struct raft_progress
 {
-    unsigned short state;      /* Probe, pipeline or snapshot. */
-    raft_index next_index;     /* Next entry to send. */
-    raft_index match_index;    /* Highest index reported as replicated. */
-    raft_index snapshot_index; /* Last index of most recent snapshot sent. */
-    raft_time last_send;       /* Timestamp of last AppendEntries RPC. */
-    bool recent_recv;          /* A msg was received within election timeout. */
+    unsigned short state;         /* Probe, pipeline or snapshot. */
+    raft_index next_index;        /* Next entry to send. */
+    raft_index match_index;       /* Highest index reported as replicated. */
+    raft_index snapshot_index;    /* Last index of most recent snapshot sent. */
+    raft_time last_send;          /* Timestamp of last AppendEntries RPC. */
+    raft_time snapshot_last_send; /* Timestamp of last InstallSnaphot RPC. */
+    bool recent_recv;             /* A msg was received within election timeout. */
 };
 
 struct raft; /* Forward declaration. */
@@ -560,16 +605,16 @@ struct raft
      *
      * The possible scenarios are:
      *
-     * 1. #configuration_index and #configuration_uncommited_index are both
+     * 1. #configuration_index and #configuration_uncommitted_index are both
      *    zero. This should only happen when a brand new server starts joining a
      *    cluster and is waiting to receive log entries from the current
      *    leader. In this case #configuration must be empty and have no servers.
      *
-     * 2. #configuration_index is non-zero while #configuration_uncommited_index
+     * 2. #configuration_index is non-zero while #configuration_uncommitted_index
      *    is zero. In this case the content of #configuration must match the one
      *    of the log entry at #configuration_index.
      *
-     * 3. #configuration_index and #configuration_uncommited_index are both
+     * 3. #configuration_index and #configuration_uncommitted_index are both
      *    non-zero, with the latter being greater than the former. In this case
      *    the content of #configuration must match the one of the log entry at
      *    #configuration_uncommitted_index.
@@ -612,6 +657,12 @@ struct raft
      */
     unsigned heartbeat_timeout;
 
+    /*
+     * When the leader sends an InstallSnapshot RPC to a follower it will consider
+     * the RPC as failed after this timeout and retry.
+     */
+    unsigned install_snapshot_timeout;
+
     /*
      * The fields below hold the part of the server's volatile state which is
      * always applicable regardless of the whether the server is follower,
@@ -694,6 +745,11 @@ struct raft
     /* Whether to use pre-vote to avoid disconnected servers disrupting the
      * current leader, as described in 4.2.3 and 9.6. */
     bool pre_vote;
+
+    /* Limit how long to wait for a stand-by to catch-up with the log when its
+     * being promoted to voter. */
+    unsigned max_catch_up_rounds;
+    unsigned max_catch_up_round_duration;
 };
 
 RAFT_API int raft_init(struct raft *r,
@@ -762,6 +818,11 @@ RAFT_API void raft_set_election_timeout(struct raft *r, unsigned msecs);
  */
 RAFT_API void raft_set_heartbeat_timeout(struct raft *r, unsigned msecs);
 
+/**
+ * Set the snapshot install timeout.
+ */
+RAFT_API void raft_set_install_snapshot_timeout(struct raft *r, unsigned msecs);
+
 /**
  * Number of outstanding log entries before starting a new snapshot. The default
  * is 1024.
@@ -781,7 +842,21 @@ RAFT_API void raft_set_pre_vote(struct raft *r, bool enabled);
 RAFT_API void raft_set_snapshot_trailing(struct raft *r, unsigned n);
 
 /**
- * Return a human-readable description of the last error occured.
+ * Set the maximum number of a catch-up rounds to try when replicating entries
+ * to a stand-by server that is being promoted to voter, before giving up and
+ * failing the configuration change. The default is 10.
+ */
+RAFT_API void raft_set_max_catch_up_rounds(struct raft *r, unsigned n);
+
+/**
+ * Set the maximum duration of a catch-up round when replicating entries to a
+ * stand-by server that is being promoted to voter. The default is 5 seconds.
+ */
+RAFT_API void raft_set_max_catch_up_round_duration(struct raft *r,
+                                                   unsigned msecs);
+
+/**
+ * Return a human-readable description of the last error occurred.
  */
 RAFT_API const char *raft_errmsg(struct raft *r);
 
@@ -897,10 +972,10 @@ RAFT_API int raft_add(struct raft *r,
  * #RAFT_BADROLE is returned.
  */
 RAFT_API int raft_assign(struct raft *r,
-			 struct raft_change *req,
-			 raft_id id,
-			 int role,
-			 raft_change_cb cb);
+                         struct raft_change *req,
+                         raft_id id,
+                         int role,
+                         raft_change_cb cb);
 
 /**
  * Remove the given server from the cluster configuration.
